import duso.business.*;
import duso.data.*;
import duso.outlierscores.LOF;

import java.util.*;
import java.io.*;

public class MainClass
{
	public static long totalTimeSortingSubspaces = 0;
	
	public static void main(String[] args)
	{
		try
		{
			readInputString(args);
			DataMatrix dataMatrix = new DataMatrix(Constants.NUM_ROWS, Constants.NUM_MEASURE_COLS);
			IndexMatrix indexMatrix = new IndexMatrix(Constants.NUM_ROWS, Constants.NUM_MEASURE_COLS);
			double[] CRES = new double[Constants.NUM_MEASURE_COLS];
			System.out.println("start reading input data...");
			DataReader.readData(Constants.FILE_INPUT, dataMatrix, indexMatrix, CRES);
			System.out.println("end reading input data...");
			System.out.println("start outlier detection...");
			outlierDetection(dataMatrix, indexMatrix, CRES);
			System.out.println("end outlier detection...");
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
	}
	
	public static void readInputString(String[] args) throws Exception
	{	
		int i;
		int total = args.length - 1;
		
		// take FILE_INPUT
		boolean found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILE_INPUT"))
			{
				Constants.FILE_INPUT = args[i + 1];
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -FILE_INPUT");
		
		// take FILE_SUB_OUTPUT
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILE_SUB_OUTPUT"))
			{
				Constants.FILE_SUB_OUTPUT = args[i + 1];
				File f = new File(Constants.FILE_SUB_OUTPUT);
				if (f.exists())
					f.delete();
				f.createNewFile();
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -FILE_SUB_OUTPUT");
		
		// take FILE_SCORES_OUTPUT
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILE_SCORES_OUTPUT"))
			{
				Constants.FILE_SCORES_OUTPUT = args[i + 1];
				File f = new File(Constants.FILE_SCORES_OUTPUT);
				if (f.exists())
					f.delete();
				f.createNewFile();
				found = true;
				break;
			}
		
		// take FILE_OSCORES_OUTPUT
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILE_OSCORES_OUTPUT"))
			{
				Constants.FILE_OSCORES_OUTPUT = args[i + 1];
				File f = new File(Constants.FILE_OSCORES_OUTPUT);
				if (f.exists())
					f.delete();
				f.createNewFile();
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -FILE_OSCORES_OUTPUT");
		
		// take NUM_ROWS
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-NUM_ROWS"))
			{
				Constants.NUM_ROWS = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -NUM_ROWS");
		
		// take NUM_MEASURE_COLS
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-NUM_MEASURE_COLS"))
			{
				Constants.NUM_MEASURE_COLS = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -NUM_MEASURE_COLS");
		
		// take FIELD_DELIMITER
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FIELD_DELIMITER"))
			{
				Constants.FIELD_DELIMITER = args[i + 1];
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -FIELD_DELIMITER");
		
		// take NUM_NEIGHBORS
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-NUM_NEIGHBORS"))
			{
				Constants.NUM_NEIGHBORS = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -NUM_NEIGHBORS");
		
		// take EPSILON
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-EPSILON"))
			{
				Constants.EPSILON = Double.parseDouble(args[i + 1]);
				found = true;
				break;
			}
		
		// take MIN_PTS
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-MIN_PTS"))
			{
				Constants.MIN_PTS = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		
		// take MAX_NUM_SUBSPACES
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-MAX_NUM_SUBSPACES"))
			{
				Constants.MAX_NUM_SUBSPACES = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -MAX_NUM_SUBSPACES");
		
		// take USE_DUSO_SEED
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-USE_DUSO_SEED"))
			{
				Constants.USE_DUSO_SEED = Boolean.parseBoolean(args[i + 1]);
				found = true;
				break;
			}
		
		// take ALPHA, RAN_INDEX_BLOCK
		if (Constants.USE_DUSO_SEED == false)
		{
			// take ALPHA
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-ALPHA"))
				{
					Constants.ALPHA = Double.parseDouble(args[i + 1]);
					found = true;
					break;
				}
			if (found == false)
				throw new Exception("Missing -ALPHA");
			
			// take RAN_INDEX_BLOCK
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-RAN_INDEX_BLOCK"))
				{
					Constants.RAN_INDEX_BLOCK = new Random(Long.parseLong(args[i + 1]));
					found = true;
					break;
				}
			if (found == false)
				Constants.RAN_INDEX_BLOCK = new Random();
		}
		
		if (Constants.SUB_EX_TYPE == Constants.NON_RANKING_SUB_EX)
		{
			// take CANDIDATE_CUTOFF
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-CANDIDATE_CUTOFF"))
				{
					Constants.CANDIDATE_CUTOFF = Integer.parseInt(args[i + 1]);
					found = true;
					break;
				}
			if (found == false)
				throw new Exception("Missing -CANDIDATE_CUTOFF");
		}
		
		// take PROB_KEEP
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-PROB_KEEP"))
			{
				Constants.PROB_KEEP = Double.parseDouble(args[i + 1]);
				found = true;
				break;
			}
		
		// take ECMI_TYPE
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-ECMI_TYPE"))
			{
				Constants.ECMI_TYPE = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
					
		if (Constants.USE_DUSO_SEED == false)
		{
			// take RAN_SUB_INDEX_BLOCK
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-RAN_SUB_INDEX_BLOCK"))
				{
					Constants.RAN_SUB_INDEX_BLOCK = new Random(Long.parseLong(args[i + 1]));
					found = true;
					break;
				}
			if (found == false)
				Constants.RAN_SUB_INDEX_BLOCK = new Random();
			
			// take SORT_ATTS
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-SORT_ATTS"))
				{
					Constants.SORT_ATTS = Integer.parseInt(args[i + 1]);
					found = true;
					break;
				}
			
			// take NUM_SUBSAMPLING
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-NUM_SUBSAMPLING"))
				{
					Constants.NUM_SUBSAMPLING = Integer.parseInt(args[i + 1]);
					found = true;
					break;
				}
			if (found == false)
				Constants.NUM_SUBSAMPLING = 1;
			
			// take PRUNE_BY_AVG
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-PRUNE_BY_AVG"))
				{
					Constants.PRUNE_BY_AVG = Boolean.parseBoolean(args[i + 1]);
					found = true;
					break;
				}
			
			// take RATIO_FROM_MEAN
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-RATIO_FROM_MEAN"))
				{
					Constants.RATIO_FROM_MEAN = Double.parseDouble(args[i + 1]);
					found = true;
					break;
				}
		}
		else
		{
			// take NUM_SEEDS
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-NUM_SEEDS"))
				{
					Constants.NUM_SEEDS = Integer.parseInt(args[i + 1]);
					found = true;
					break;
				}
			if (found == false)
				throw new Exception("Missing -NUM_SEEDS");
			
			// take RAN_SEED_GEN
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-RAN_SEED_GEN"))
				{
					Constants.RAN_SEED_GEN = new Random(Long.parseLong(args[i + 1]));
					found = true;
					break;
				}
			if (found == false)
				Constants.RAN_SEED_GEN = new Random();
			
			// take UPDATE_SEED
			found = false;
			for (i = 0; i < total; i++)
				if (args[i].equals("-UPDATE_SEED"))
				{
					Constants.UPDATE_SEED = Boolean.parseBoolean(args[i + 1]);
					found = true;
					break;
				}
		}
	}
	
	public static void outlierDetection(DataMatrix dataMatrix, IndexMatrix indexMatrix, double[] CRES) throws Exception
	{
		try
		{
			ArrayList<Subspace> interestingSubspaces = null;
			ArrayList<Double> scores;
			BufferedWriter writerSub = new BufferedWriter(new FileWriter(new File(Constants.FILE_SUB_OUTPUT)));
			BufferedWriter writerOverallScores = new BufferedWriter(new FileWriter(new File(Constants.FILE_OSCORES_OUTPUT)));
			
			int numSubspaces = -1;
			int rows = dataMatrix.rows;
			double[] overallScores = new double[rows];
			if (Constants.SCORE_INT_TYPE == Constants.PROD)
			{
				for (int i = 0; i < rows; i++)
					overallScores[i] = 1;
			}
			
			System.out.println("start extracting interesting subspaces...");
			if (Constants.SUB_EX_TYPE == Constants.NON_RANKING_SUB_EX)
				interestingSubspaces = SubspaceExploration.exploreSubspaces(dataMatrix, indexMatrix, Constants.CANDIDATE_CUTOFF, CRES, writerSub);
			System.out.println("end extracting interesting subspaces...");
			if (interestingSubspaces != null && interestingSubspaces.size() > Constants.MAX_NUM_SUBSPACES)
				interestingSubspaces = getTopSubspaces(interestingSubspaces);
			System.out.println("start computing outlier scores...");
			Subspace tmpSpace = null;
			int numDims;
			if (interestingSubspaces != null)
				numSubspaces = interestingSubspaces.size();
			
			for (int i = 0; i < numSubspaces; i++)
			{
				tmpSpace = interestingSubspaces.get(i);
				numDims = tmpSpace.num_dims;
				
				for (int j = 0; j < numDims; j++)
				{
					if (j == 0)
						writerSub.write(Integer.toString(tmpSpace.dims.get(j).intValue()));
					else
						writerSub.write("," + Integer.toString(tmpSpace.dims.get(j).intValue()));
				}
				writerSub.newLine();
				
				if (Constants.SCORE_TYPE == Constants.LOF || Constants.SCORE_TYPE == Constants.KNN)
				{
					scores = null;
					if (Constants.SCORE_TYPE == Constants.LOF)
						scores = LOF.computeLOF(dataMatrix, tmpSpace);
					
					if (scores != null)
					{
						if (Constants.SCORE_INT_TYPE == Constants.SUM)
						{
							for (int j = 0; j < rows; j++)
								overallScores[j] += scores.get(j).doubleValue();
						}
						else if (Constants.SCORE_INT_TYPE == Constants.PROD)
						{
							for (int j = 0; j < rows; j++)
								overallScores[j] *= scores.get(j).doubleValue();
						}
					}
				}
				
			} // end for
			System.out.println("end computing outlier scores...");
			
			if (Constants.SCORE_TYPE == Constants.LOF || Constants.SCORE_TYPE == Constants.KNN)
			{
				for (int i = 0; i < rows; i++)
				{
					writerOverallScores.write(Integer.toString(i) + " " + Double.toString(overallScores[i]) + " " + dataMatrix.data.get(i).classID);
					writerOverallScores.newLine();
				}
			}
			
			writerSub.flush();
			writerSub.close();
			writerOverallScores.flush();
			writerOverallScores.close();
		}
		catch (Exception ex)
		{
			throw ex;
		}
	} // end method
	
	public static ArrayList<Subspace> getTopSubspaces(ArrayList<Subspace> candidateSubspaces)
	{
		ArrayList<Subspace> ret = new ArrayList<Subspace>();
		int size = candidateSubspaces.size();
		
		SortedObjectComparator c = new SortedObjectComparator();
		SortedObject[] tmp = new SortedObject[size];
		for (int i = 0; i < size; i++)
		{
			tmp[i] = new SortedObject(0, 0);
			tmp[i].index = i;
			tmp[i].value = candidateSubspaces.get(i).contrast;
		}
		Arrays.sort(tmp, c);
		
		int upper = size - 1;
		int lower = size - Constants.MAX_NUM_SUBSPACES;
		for (int i = upper; i >= lower; i--)
			ret.add(candidateSubspaces.get(tmp[i].index));
		
		return ret;
	}
}
