package duso.business;

import duso.data.*;

import java.io.*;
import java.util.*;

public class DataReader 
{
	public static void readData(String fileName, DataMatrix dataMatrix, IndexMatrix indexMatrix, double[] CRES) throws Exception
	{
		try
		{
			Scanner scanner = new Scanner(new FileInputStream(fileName));
			
			String s = null;
			String[] vals = null;
			int rows = dataMatrix.rows;
			int cols = dataMatrix.cols;
			DataPoint newPoint = null;
			int count = 0;
			
			double[] maxDims = new double[cols];
			double[] minDims = new double[cols];
			Constants.DATA_MEANS  = new double[cols];
			Constants.DATA_DEVS  = new double[cols];
			for (int i = 0; i < cols; i++)
			{
				maxDims[i] = -Double.MAX_VALUE;
				minDims[i] = Double.MAX_VALUE;
				Constants.DATA_MEANS[i] = 0;
				Constants.DATA_DEVS[i] = 0;
			}
			
			while (scanner.hasNextLine())
			{
				s = scanner.nextLine();
	        	vals = s.split(Constants.FIELD_DELIMITER);
	        	
	        	if (vals.length != cols + 1)
	        		throw new Exception("Invalid input!");
	        	
		        newPoint = new DataPoint(cols, 0, 0);
		        newPoint.universalID = count;
		        newPoint.classID = Integer.parseInt(vals[cols]);
		        for (int j = 0; j < cols; j++)
		        {
		        	vals[j] = vals[j].replace(',', '.');
		        	newPoint.measures[j] = Double.parseDouble(vals[j]);
		        	
		        	if (newPoint.measures[j] > maxDims[j])
		        		maxDims[j] = newPoint.measures[j];
		        	
		        	if (newPoint.measures[j] < minDims[j])
		        		minDims[j] = newPoint.measures[j];
		        	
		        	Constants.DATA_MEANS[j] += newPoint.measures[j] / rows;
		        }
		        
		        dataMatrix.data.add(newPoint);
		        
		        count++;
		        if (count == rows)
		        	break;
		    } // end while
			
			double[] dimData = null;
			SortedObjectComparator c = new SortedObjectComparator();
			SortedObject[] tmp = new SortedObject[cols];
			for (int i = 0; i < cols; i++)
			{
				dimData = new double[rows];
				for (int j = 0; j < rows; j++)
				{
					dimData[j] = dataMatrix.data.get(j).measures[i];
					Constants.DATA_DEVS[i] += (dimData[j] - Constants.DATA_MEANS[i]) * (dimData[j] - Constants.DATA_MEANS[i]) / (rows - 1);
				}
				Constants.DATA_DEVS[i] = Math.sqrt(Constants.DATA_DEVS[i]);
				
				CRES[i] = DUSOProcessing.computeCRE(dimData, false, Constants.USE_RAO_CRE);

				tmp[i] = new SortedObject(0, 0);
				tmp[i].index = i;
				tmp[i].value = CRES[i];
			}
			
			if (Constants.SORT_ATTS == Constants.SORT_ATTS_ASC || Constants.SORT_ATTS == Constants.SORT_ATTS_DSC)
			{
				Arrays.sort(tmp, c);
				double[] tmpMeasures = new double[cols];
				DataPoint curPoint = null;
				
				if (Constants.SORT_ATTS == Constants.SORT_ATTS_ASC)
				{
					for (int i = 0; i < rows; i++)
					{
						curPoint = dataMatrix.data.get(i);
						for (int j = 0; j < cols; j++)
						{
							tmpMeasures[j] = curPoint.measures[tmp[j].index];
							if (i == 0)
								CRES[j] = tmp[j].value;
						}
					
						for (int j = 0; j < cols; j++)
							curPoint.measures[j] = tmpMeasures[j];
					} // end for
				}
				else
				{
					for (int i = 0; i < rows; i++)
					{
						curPoint = dataMatrix.data.get(i);
						for (int j = cols - 1; j >= 0; j--)
						{
							tmpMeasures[cols - 1 - j] = curPoint.measures[tmp[j].index];
							if (i == 0)
								CRES[cols - 1 - j] = tmp[j].value;
						}
					
						for (int j = 0; j < cols; j++)
							curPoint.measures[j] = tmpMeasures[j];
					} // end for
				} // end if
			} // end if
			
			DataProcess.calculateIndices(dataMatrix, indexMatrix);
			scanner.close();
		}
		catch (Exception ex)
		{
			throw ex;
		}
	}	
}
