package duso.business;

import duso.data.*;

import java.io.BufferedWriter;
import java.util.*;

class DataWithIndex
{
	public DataMatrix data;
	public IndexMatrix index;
	public ArrayList<double[][]> sortedData;
}

public class SubspaceExploration 
{
	public static ArrayList<Subspace> exploreSubspaces(DataMatrix dataMatrix, IndexMatrix indexMatrix, int candidateCutOff, double[] CRES, BufferedWriter writerSub) throws Exception
	{
		ArrayList<Subspace> ret = new ArrayList<Subspace>();
		int totalCols = dataMatrix.cols;
		ArrayList<Subspace> curCandidates = new ArrayList<Subspace>();
		
		Subspace tmp = null;
		for (int i = 0; i < totalCols; i++)
			for (int j = i + 1; j < totalCols; j++)
			{
				tmp = new Subspace(2);
				tmp.dims.add(new Integer(i));
				tmp.dims.add(new Integer(j));
				curCandidates.add(tmp);
			}
		
		DataWithIndex orgData = new DataWithIndex();
		orgData.data = dataMatrix;
		orgData.index = indexMatrix;
		
		ArrayList<Integer> bestCandidateID = null;
		
		int curSubspaceDim = 2;
		while (curCandidates.size() > 0 && curSubspaceDim <= Constants.NUM_MEASURE_COLS)
		{
			bestCandidateID = new ArrayList<Integer>();
			int curSize = curCandidates.size();
			DataWithIndex curData = null;
			
			Subspace tmpSubspace = null;
			for (int i = 0; i < curSize; i++)
			{
				tmpSubspace = curCandidates.get(i);
				curData = sliceCols(orgData, tmpSubspace.dims);
				
				Subspace.computeSubspaceContrast(tmpSubspace, CRES, curData.data, curData.index, curData.sortedData);
				
				if (Constants.PRUNE_BY_AVG == false || tmpSubspace.contrast > Constants.UPPER_VAL * (tmpSubspace.num_dims - 1) * Constants.RATIO_FROM_MEAN)
				{
					int numBestCandidates = bestCandidateID.size();
					Subspace curBestSubspace = null;
					int pos = numBestCandidates;
					for (int j = numBestCandidates - 1; j >= 0; j--)
					{
						curBestSubspace = curCandidates.get(bestCandidateID.get(j).intValue());
						if (tmpSubspace.contrast <= curBestSubspace.contrast)
							break;
						
						pos = j;
					} // end for
					
					bestCandidateID.add(pos, new Integer(i));
					numBestCandidates++;
				
					if (Constants.PRUNE_BY_AVG == false)
					{
						if (numBestCandidates > candidateCutOff)
							for (int j = numBestCandidates - 1; j >= candidateCutOff; j--)
								bestCandidateID.remove(j);
					}
				}
			} // end for
			
			int size = bestCandidateID.size();
			ArrayList<Subspace> retainedSubspaces = new ArrayList<Subspace>(); // contain list of subspaces to retain for further subspace generation
			Subspace retainedSub = null;
			Subspace resultSub = null;
			Subspace can = null;
			for (int i = 0; i < size; i++)
			{
				retainedSub = new Subspace();
				can = curCandidates.get(bestCandidateID.get(i).intValue());
				retainedSub.dims = can.dims;
				retainedSub.num_dims = can.num_dims;
				retainedSub.contrast = can.contrast;
				
				resultSub = new Subspace();
				resultSub.dims = can.dims;
				resultSub.num_dims = can.num_dims;
				resultSub.contrast = can.contrast;
				
				retainedSubspaces.add(retainedSub);
				ret.add(resultSub);
				
				int retSize = ret.size();
				int start = retSize - 1;
				while (start >= 0 && ret.get(start).num_dims != resultSub.num_dims - 1) // find the first subspace having number of dimensions less than the current subspace
					start--;
				for (int j = start; j >= 0; j--)
				{
					Subspace s = ret.get(j);
					if (s.num_dims != resultSub.num_dims - 1)
						break;
					
					if (Subspace.isParentSubspace(resultSub, s))
						if (resultSub.contrast >= s.contrast)
							ret.remove(j); // check if there is bug at this line
				}
			}
			
			curCandidates = generateCandidateSubspaces(retainedSubspaces, Constants.PRUNE_NON_CAND_SUB);
			curSubspaceDim++;
			//System.out.println("number of candidate subspaces: " + curCandidates.size());
		} // end while
		
		return ret;
	}
	
	public static ArrayList<Subspace> generateCandidateSubspaces(ArrayList<Subspace> sourceSubspaces, boolean pruneNonCandidate)
	{
		ArrayList<Subspace> ret = new ArrayList<Subspace>();
		int numSource = sourceSubspaces.size();
		Subspace tmp = null;
		for (int i = 0; i < numSource; i++)
			for (int j = i + 1; j < numSource; j++)
			{
				tmp = Subspace.combineSubspaces(sourceSubspaces.get(i), sourceSubspaces.get(j));
				if (tmp != null)
				{
					boolean exists = false;
					int curRetSize = ret.size();
					for (int k = 0; k < curRetSize; k++)
						if (Subspace.isIndenticalSubspaces(tmp, ret.get(k)))
						{
							exists = true;
							break;
						}
					
					if (exists == false)
					{
						if (pruneNonCandidate)
						{
							if (!Subspace.hasNonCandidateSubspaces(tmp, sourceSubspaces))
								ret.add(tmp);
							else
							{
								double prob = Constants.RAN_KEEP.nextDouble();
								if (prob < Constants.PROB_KEEP)
									ret.add(tmp);
							}
						}
						else
							ret.add(tmp);
					} // end if
				} // end if
			} // end for
		
		return ret;
	}
	
	public static DataWithIndex sliceCols(DataWithIndex di, ArrayList<Integer> dims)
	{
		int rows = di.data.rows;
		int newCols = dims.size();
		
		DataWithIndex ret = new DataWithIndex();
		ret.data = new DataMatrix(rows, newCols);
		ret.index = new IndexMatrix(rows, newCols);
		
		DataPoint curPoint = null;
		DataPoint tmpPoint = null;
		int val;
		for (int i = 0; i < rows; i++)
		{
			curPoint = di.data.data.get(i);
			tmpPoint = new DataPoint(newCols, 0, 0);
			for (int j = 0; j < newCols; j++)
			{
				val = dims.get(j).intValue();
				tmpPoint.measures[j] = curPoint.measures[val];
				ret.index.data[i][j] = di.index.data[i][val];
			}
			
			ret.data.data.add(tmpPoint);
		}
		
		return ret;
	}
	
	public static ArrayList<Subspace> getRandomSubspaces(DataMatrix dataMatrix, IndexMatrix indexMatrix, int numDims, int numSubspaces, Random dimGenerator, Random subspaceIndexBlock)
	{
		DataWithIndex orgData = new DataWithIndex();
		orgData.data = dataMatrix;
		orgData.index = indexMatrix;
				
		ArrayList<Subspace> ret = new ArrayList<Subspace>();
		Subspace tmpSubspace = null;
		int lowerBound = numDims / 2;
		int upperBound = numDims - 1;
		int subspaceDim;
		int maxPos;
		int pos;
		int[] tmpDims = null;
		
		ArrayList<Integer> dims = new ArrayList<Integer>();
		for (int i = 0; i < numDims; i++)
			dims.add(new Integer(i));
		
		for (int i = 0; i < numSubspaces; i++)
		{
			Collections.shuffle(dims);
			subspaceDim = DataProcess.generateRandomNumber(dimGenerator, lowerBound, upperBound, false);
			maxPos = numDims - subspaceDim;
			pos = DataProcess.generateRandomNumber(subspaceIndexBlock, 0, maxPos, false);
			tmpSubspace = new Subspace(subspaceDim);
			tmpDims = new int[subspaceDim];
			for (int j = 0; j < subspaceDim; j++)
				tmpDims[j] = pos + j;
			
			Arrays.sort(tmpDims);
			for (int j = 0; j < subspaceDim; j++)
				tmpSubspace.dims.add(new Integer(tmpDims[j]));
			
			ret.add(tmpSubspace);
		}
		
		return ret;
	}
	
	public static ArrayList<Subspace> getOneRandomSubspace(int totalNumDims, int numDims)
	{
		ArrayList<Subspace> ret = new ArrayList<Subspace>();
		Subspace tmpSubspace = new Subspace(numDims);
		int start = totalNumDims - numDims;
		for (int i = start; i < totalNumDims; i++)
			tmpSubspace.dims.add(new Integer(i));
		
		ret.add(tmpSubspace);
		return ret;
	}
}