package duso.data;

import java.util.*;

import duso.business.*;

public class Subspace 
{
	public int num_dims;
	public ArrayList<Integer> dims;
	public double contrast;
	public double entropy;
	
	public Subspace()
	{
		// do nothing
	}
	
	public Subspace(int num_dims)
	{
		this.num_dims = num_dims;
		dims = new ArrayList<Integer>();
	}
	
	public static void computeSubspaceContrast(Subspace S, double[] CRES, DataMatrix dataMatrix, IndexMatrix indexMatrix, ArrayList<double[][]> sortedData) throws Exception
	{
		ArrayList<Integer> dims = new ArrayList<Integer>();
		double[] usedCRES = new double[S.num_dims];
		double[] usedMeans = new double[S.num_dims];
		double[] usedDevs = new double[S.num_dims];
		for (int i = 0; i < S.num_dims; i++)
		{
			dims.add(new Integer(i));
			usedCRES[i] = CRES[S.dims.get(i).intValue()];
			usedMeans[i] = Constants.DATA_MEANS[S.dims.get(i).intValue()];
			usedDevs[i] = Constants.DATA_DEVS[S.dims.get(i).intValue()];
		}
		
		S.contrast = DUSOProcessing.computeNormalizedCMOIMI(dataMatrix, indexMatrix, dims, usedCRES, Constants.ALPHA, Constants.RAN_INDEX_BLOCK, Constants.RAN_SUB_INDEX_BLOCK, Constants.NORM_CONSTRAINT, Constants.NORM_TYPE, Constants.UPPER_VAL, usedMeans, usedDevs);
	}
	
	public static boolean isIndenticalSubspaces(Subspace S, Subspace T)
	{
		int dimS = S.num_dims;
		int dimT = T.num_dims;
		
		if (dimS != dimT)
			return false;
		
		for (int i = 0; i < dimS; i++)
			if (S.dims.get(i).intValue() != T.dims.get(i).intValue())
				return false;
		
		return true;
	}
	
	public static boolean isParentSubspace(Subspace S, Subspace T)
	{
		int dimS = S.num_dims;
		int dimT = T.num_dims;
		
		if (dimT >= dimS)
			return false;
		
		for (int i = 0; i < dimT; i++)
			if (!S.dims.contains(T.dims.get(i)))
				return false;
		
		return true;
	}
	
	public static boolean hasNonCandidateSubspaces(Subspace S, ArrayList<Subspace> source)
	{
		int dim = S.num_dims;
		int sourceSize = source.size();
		ArrayList<Integer> tmpSub = new ArrayList<Integer>();
		for (int i = 0; i < dim; i++)
			tmpSub.add(new Integer(S.dims.get(i).intValue()));
		
		Subspace sub = new Subspace();
		sub.num_dims = dim;
		sub.dims = tmpSub;
		boolean ret = false;
		boolean exists;
		Integer removedDim = null;
		for (int i = 0; i < dim; i++)
		{
			removedDim = sub.dims.remove(i);
			sub.num_dims--;
			exists = false;
			for (int j = 0; j < sourceSize; j++)
				if (isIndenticalSubspaces(sub, source.get(j)))
				{
					exists = true;
					break;
				}
			
			if (exists == false)
			{
				ret = true;
				break;
			}
			
			sub.dims.add(i, removedDim);
			sub.num_dims++;
		}
		
		return ret;
	}
	
	public static Subspace combineSubspaces(Subspace S, Subspace T)
	{
		int dimS = S.num_dims;
		int dimT = T.num_dims;
		
		if (dimS != dimT)
			return null;
		
		ArrayList<Integer> tmpDimS = new ArrayList<Integer>();
		ArrayList<Integer> tmpDimT = new ArrayList<Integer>();
		for (int i = 0; i < dimS; i++)
		{
			tmpDimS.add(new Integer(S.dims.get(i).intValue()));
			tmpDimT.add(new Integer(T.dims.get(i).intValue()));
		}
		
		tmpDimT.removeAll(tmpDimS);
		if (tmpDimT.size() == 1)
		{
			int newDim = tmpDimT.get(0);
			int pos = dimS;
			for (int i = dimS - 1; i >= 0; i--)
			{
				if (newDim > tmpDimS.get(i).intValue())
					break;
				
				pos = i;
			}
			
			tmpDimS.add(pos, new Integer(newDim));
			Subspace newSubspace = new Subspace();
			newSubspace.num_dims = tmpDimS.size();
			newSubspace.dims = tmpDimS;
			return newSubspace;
		}
		else
			return null;
	}
}
