package duso.outlierscores;

import duso.business.Constants;
import duso.data.*;
import java.util.*;

public class LOF 
{
	public static ArrayList<Double> computeLOF(DataMatrix dataMatrix, Subspace S)
	{
		ArrayList<Double> ret = new ArrayList<Double>();
		computeNearestNeighbors(dataMatrix, S);
		computeLRD(dataMatrix, S);
		DataPoint curPoint = null;
		int numNeighbors;
		double sum;
		for (int i = 0; i < dataMatrix.rows; i++)
		{
			curPoint = dataMatrix.data.get(i);
			numNeighbors = curPoint.neighbors.size();
			sum = 0;
			for (int j = 0; j < numNeighbors; j++)
				sum += dataMatrix.data.get(curPoint.neighbors.get(j).intValue()).lrd;
			
			if (curPoint.lrd == 0 || Double.isNaN(curPoint.lrd) || Double.isInfinite(curPoint.lrd) || Double.isNaN(sum) || Double.isInfinite(sum))
				return null;
			
			ret.add(new Double(sum / (numNeighbors * curPoint.lrd)));
		}
		
		return ret;
	}
	
	public static void computeLRD(DataMatrix dataMatrix, Subspace S)
	{
		DataPoint curPoint = null;
		int numNeighbors;
		double sum;
		for (int i = 0; i < dataMatrix.rows; i++)
		{
			curPoint = dataMatrix.data.get(i);
			numNeighbors = curPoint.neighbors.size();
			sum = 0;
			for (int j = 0; j < numNeighbors; j++)
				sum += reachabilityDistance(S, curPoint, dataMatrix.data.get(curPoint.neighbors.get(j).intValue()));
			
			curPoint.lrd = numNeighbors / sum;
		}
	}
	
	public static void computeNearestNeighbors(DataMatrix dataMatrix, Subspace S)
	{
		DataPoint curPoint = null;
		for (int i = 0; i < dataMatrix.rows; i++)
		{
			curPoint = dataMatrix.data.get(i);
			curPoint.neighbors.clear();
			curPoint.distToNeighbors.clear();
			curPoint.lrd = 0;
			for (int j = 0; j < dataMatrix.rows; j++)
				if (j != i)
					curPoint.insertNeighbors(S, dataMatrix.data.get(j));
		}
	}
	
	public static double reachabilityDistance(Subspace S, DataPoint a, DataPoint b)
	{
		double bKthDist = b.distToNeighbors.get(Constants.NUM_NEIGHBORS - 1).doubleValue();
		double dist = DataPoint.distanceLNorm(Constants.LP_NORM, S, a, b);
		return Math.max(bKthDist, dist);
	}
}
